/*********************************************************************/
/* main.c                                                            */
/* a bad example of a storage/retrieval application                  */
/* using neblo as a routing layer.                                   */
/* see COPYRIGHT in parent directory                                 */
/*********************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <sys/time.h>
#include <openssl/rand.h>

#include <neblo.h>

#define MAX_LIFE_SECS 240  /* peer lifetime is unif. random on [0..240[ sec */
#define ULL unsigned long long
#define CP 3

#define min(a,b) (a<b?a:b)

typedef struct rt {
  neblo_addr addr;
  char data[NEBLO_MSGSIZE];
  unsigned int len;
  struct rt *next;
} record_t;

record_t    *record_head = NULL, *record_tail = NULL;

neblo_port   rport;
neblo_port   wport;


int
w_handler
( neblo_port p, neblo_addr a, void *msg, unsigned int *len )
{

  record_t    *r;


  for (r = record_head; r != NULL; r = r->next)
    if (r->addr == a)
      break;
  if (r != NULL) {
    fprintf(stderr,"w_handler(): addr %016llx already found, no rewrite\n",
            (ULL)a);
    *len = 0;  /* empty response back */
    return 0;
  }
  r = (record_t *)calloc(1,sizeof(record_t));
  if (r == NULL) {
    fprintf(stderr,"w_handler(): calloc failed\n");
    return -1;
  }

  r->addr = a;
  memcpy(r->data,msg,*len);
  r->len = *len;
  r->next = NULL;
  if (record_head == NULL) {
    record_head = r;
    record_tail = r;
  }
  else {
    record_tail->next = r;
    record_tail = r;
  }
  fprintf(stderr,"w_handler(): wrote addr %016llx size %u msg %s\n",
                  (ULL)a,*len,(char *)msg);

  *len = 0;  /* empty response back */
  return 0;

}


int
r_handler
( neblo_port p, neblo_addr a, void *msg, unsigned int *len )
{

  record_t    *r;


  for (r = record_head; r != NULL; r = r->next)
    if (r->addr == a)
      break;
  if (r == NULL) {
    fprintf(stderr,"r_handler(): addr %016llx NOT FOUND\n",(ULL)a);
    return -1;
  }

  memcpy(msg,r->data,r->len);
  *len = r->len;
  fprintf(stderr,"r_handler(): found addr %016llx size %u msg %s\n",
          (ULL)a,*len,(char *)msg);

  return 0;

}


int
refresh_handler
( neblo_addr lo_addr, neblo_addr up_addr )
{

  record_t    *r;
  neblo_addr   prev_addr = 0;
  int          rc;


  fprintf(stderr,
          "refresh_handler(): intvl %016llx -- %016llx\n",
          (ULL)lo_addr,(ULL)up_addr);
  for (r = record_head; r != NULL; r = r->next) {
    if (neblo_in_range(r->addr,lo_addr,up_addr))
      if (r->addr != prev_addr || r == record_head) {
        fprintf(stderr,"refresh_handler(): block %016llx...\n",(ULL)r->addr);
        rc = neblo_UD_send_h(wport,r->addr,r->data,r->len);
        if (rc < 0)
          fprintf(stderr,"FAILED (delayed) (status %d)\n",rc);
        else
          fprintf(stderr,"OK\n");
        prev_addr = r->addr;
      }
  }

  return 0;

}


int
clean_handler
( neblo_addr lo_addr, neblo_addr up_addr )
{

  record_t    *r;
  record_t    *rr;


  fprintf(stderr,"clean_handler(): intvl %016llx -- %016llx\n",
          (ULL)lo_addr,(ULL)up_addr);
  for (r = rr = record_head; r != NULL;) {
    if (neblo_in_range(r->addr,lo_addr,up_addr)) {
      fprintf(stderr,"clean_handler(): clean block %016llx\n",(ULL)r->addr);
      if (rr == r) {
        rr = r->next;
        free(r);
        r = rr;
        record_head = rr;
      }
      else {
        rr->next = r->next;
        free(r);
        r = rr->next;
      }
    }
    else {
      rr = r;
      r = r->next;
    }
  }
  record_tail = rr;

  return 0;

}


int 
main( int argc, char **argv )
{
  
  double                val;
  char                 *send, *recv;
  void                **vector;
  neblo_addr            addr;
  neblo_rhandle         h;
  int                   rc;
  int                   msgs = 0;
  int                   killer = 0;
  unsigned int          i;
  unsigned int          life_secs;
  struct timeval        tv,tvstart;


  /* init the random generator */
  gettimeofday(&tv,NULL);
  srand(tv.tv_usec);
  do {
    val = rand();
    RAND_seed((void *)&val,sizeof(val));
  } while (!RAND_status());

  send = malloc(NEBLO_MSGSIZE);
  recv = malloc(NEBLO_MSGSIZE);
  vector = malloc(CP*sizeof(char *));

  neblo_init(argc,argv);

  rport = 101;
  wport = 102;

  neblo_set_receiver_handler_REQ(wport,w_handler);
  neblo_set_receiver_handler_UD(wport,w_handler);
  neblo_set_receiver_handler_REQ(rport,r_handler);
  /*  No forwarder handlers in this small application.  */
  neblo_set_refresh_handler(refresh_handler);
  neblo_set_clean_handler(clean_handler);

  neblo_connect();

  for (;;) {
    while ((rc = getopt(argc,argv,"dm")) != 'd' && rc != 'm' && rc != -1);
    if (rc == -1)
      break;
    switch (rc) {
      case 'd':
        killer = 1;
        break;
      case 'm':
        msgs = 1;
        break;
    }
  }

  life_secs = (int)(MAX_LIFE_SECS*(double)rand()/(double)RAND_MAX);

  if (msgs)
    getchar();

  gettimeofday(&tvstart,NULL);

  for (;;) {
    sleep(1);
    if (killer) {
      gettimeofday(&tv,NULL);
      if ((unsigned int)(tv.tv_sec - tvstart.tv_sec) >= life_secs)
        /* lifetime expired */
        break;
    }
    if (!msgs)
      continue;
    RAND_bytes((unsigned char *)&addr,sizeof(addr));
    sprintf(send,"AA %u OOOO %u ZZ\n",rand(),rand());
    for (i=0;i<CP;i++)
      vector[i] = NULL;
    for (i=3; i; i--) {
      fprintf(stderr,"store to addr %016llx, attempt %u...\n",(ULL)addr,3-i+1);
      neblo_REQ_send(CP,wport,addr,send,strlen(send),vector,0,&h);
      rc = neblo_RES_wait(h,NEBLO_WAITALL);
      if (rc >= 0) break;
    }
    if (!i) {
      fprintf(stderr,"FAILURE did not send data\n");
      continue;
    }
    memset(recv,0,NEBLO_MSGSIZE);
    for (i=0;i<CP;i++)
      vector[i] = recv;
    sleep(1);
    for (i=3; i; i--) {
      fprintf(stderr,"retrieve, attempt %u...\n",3-i+1);
      neblo_REQ_send(CP,rport,addr,NULL,0,vector,NEBLO_MSGSIZE,&h);
      rc = neblo_RES_wait(h,NEBLO_WAITANY);
      if (rc >= 0) break;
    }
    if (!i)
      fprintf(stderr,"FAILURE did not retrieve data\n");
    else
      if (strcmp(send,recv))
        fprintf(stderr,"FAILURE data do not match (rc %d, h %u): "
                "got %s (len %u,%u %u %u) instead of %s (len %u, %u %u %u)\n",
rc,(unsigned int)h,
recv,strlen(recv),
recv[strlen(recv)-3],
recv[strlen(recv)-2],
recv[strlen(recv)-1],
send,strlen(send),
send[strlen(send)-3],
send[strlen(send)-2],
send[strlen(send)-1]
);
      else
        fprintf(stderr,"SUCCESS!!!!\n");
  }

  return 0;

}
