#ifndef NEBLO_SECRET_H

#define NEBLO_SECRET_H

#include <openssl/rsa.h>                /* public key encryption        */

#define DEFAULTPUBKEYFILE "neblo.pk"
#define DEFAULTSECKEYFILE "neblo.sk"
#define DEFAULTKEYFILE "neblo"
#define DEFAULTBITKEYSIZE 2048
#define DEFAULTBYTEKEYSIZE (DEFAULTBITKEYSIZE/8)
#define MAXBITKEYSIZE 4096
#define MAXBYTEKEYSIZE (MAXBITKEYSIZE/8)
#define SESSIONKEYSIZE 32               /* 32 byte = 256 bit session keys */
#define DEFAULTKEYEXP 65357

#define _convert_pubkey_(rsapt,pubkeymod,pubkeyexp,cont)        \
{                                                               \
  BN_hex2bn(&((rsapt)->n), pubkeymod);                          \
  BN_hex2bn(&((rsapt)->e), pubkeyexp);                          \
  if (!((rsapt)->n && (rsapt)->e)) {                            \
    cont;                                                       \
  }                                                             \
}

#define _get_public_key_(fp, rsapt, buf)                \
do {                                                    \
  rewind(fp);						\
  fscanf(fp, "%s", buf);				\
  BN_hex2bn(&(rsapt->n), buf);				\
  fscanf(fp, "%s", buf);				\
  BN_hex2bn(&(rsapt->e), buf);				\
} while (0)

#define _get_private_key_(fp, rsapt, buf)		\
do {                                                    \
  (rsapt) = RSA_new();                                  \
  if (!(rsapt)) {                                       \
    perror("_get_private_key_: RSA_new() failed");      \
    exit(1);                                            \
  }                                                     \
  rewind(fp);						\
  (buf)[0] = 0;                                         \
  fscanf(fp, "%s", buf);				\
  BN_hex2bn(&(rsapt->n), buf);				\
  fscanf(fp, "%s", buf);				\
  BN_hex2bn(&(rsapt->e), buf);				\
  fscanf(fp, "%s", buf);				\
  BN_hex2bn(&(rsapt->d), buf);				\
  fscanf(fp, "%s", buf);				\
  BN_hex2bn(&(rsapt->p), buf);				\
  fscanf(fp, "%s", buf);				\
  BN_hex2bn(&(rsapt->q), buf);				\
  fscanf(fp, "%s", buf);				\
  BN_hex2bn(&(rsapt->dmp1), buf);			\
  fscanf(fp, "%s", buf);				\
  BN_hex2bn(&(rsapt->dmq1), buf);			\
  fscanf(fp, "%s", buf);				\
  BN_hex2bn(&(rsapt->iqmp), buf);			\
} while (0)

#define _clean_private_key_(rsapt, buf, buflen)         \
do {                                                    \
  memset(buf,0,buflen);                                 \
  RSA_free(rsapt);                                      \
} while (0)

#define _RSA_new_(rsapt) \
{\
  pthread_mutex_lock(&malloc_lock);\
  rsapt = RSA_new();\
  pthread_mutex_unlock(&malloc_lock);\
}

#define _RSA_free_(rsapt) \
{\
  pthread_mutex_lock(&malloc_lock);\
  if ((rsapt) && (rsapt)->n && (rsapt)->e) {\
    RSA_free(rsapt);\
    rsapt = NULL;\
  } \
  pthread_mutex_unlock(&malloc_lock);\
}

#define _RSA_copy_(rsaptdest,rsaptsrc,cont) \
{\
  pthread_mutex_lock(&malloc_lock);\
  rsaptdest = RSA_new();\
  (rsaptdest)->n = BN_dup((rsaptsrc)->n);\
  (rsaptdest)->e = BN_dup((rsaptsrc)->e);\
  pthread_mutex_unlock(&malloc_lock);\
  if (!((rsaptdest)->n && (rsaptdest)->e)) {\
    _RSA_free_(rsaptdest);\
    cont;\
  }\
}

#define _BN_bn2hex_(buf,bignum) \
{\
  pthread_mutex_lock(&malloc_lock);\
  buf = BN_bn2hex(bignum);\
  pthread_mutex_unlock(&malloc_lock);\
}

#define _OPENSSL_free_(buf) \
{\
  pthread_mutex_lock(&malloc_lock);\
  if (buf) {\
    OPENSSL_free(buf);\
    buf = NULL;\
  } \
  pthread_mutex_unlock(&malloc_lock);\
}


#endif
